/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        hiscore.cpp ( 7DRL Library, C++ )
 *
 * COMMENTS:
 */

#include "hiscore.h"
#include "ptrlist.h"
#include "msg.h"
#include "grammar.h"
#include "text.h"

#include "glbdef.h"

#include <fstream>
#include <iostream>
using namespace std;

PTRLIST<BUF>	theScoreFiles;
PTRLIST<int>	theScoreValues;

void
hiscore_load()
{
    char		line[100];
    int			idx, score;
    ifstream		is("../rooms/highscore.txt");
    BUF			mapname;

    theScoreFiles.clear();
    theScoreValues.clear();

    while (is.getline(line, 100))
    {
	text_striplf(line);

	// Ignore comments.
	if (line[0] == '#')
	    continue;
	
	mapname.clear();
	// Find the score, : prefixed.
	for (idx = 0; line[idx] && line[idx] != ':'; idx++)
	{
	    mapname.append(line[idx]);
	}

	if (!line[idx])
	{
	    // No : component, invalid score.
	    continue;
	}

	// Skip any spaces...
	idx++;
	while (ISSPACE(line[idx]) && line[idx])
	    idx++;

	// Extract our score.
	score = atoi(&line[idx]);

	// Add to our score list.
	theScoreFiles.append(mapname);
	theScoreValues.append(score);
    }
}

void
hiscore_save()
{
    char		line[100];
    int			idx, score;
    ofstream		os("../rooms/highscore.txt");
    BUF			mapname;

    // Output our comments.
    os << "# High Score file for Fatherhood" << endl;
    os << endl;

    for (idx = 0; idx < theScoreFiles.entries(); idx++)
    {
	os << theScoreFiles(idx).buffer() << ": " << theScoreValues(idx) << endl;
    }
}

void
hiscore_display()
{
    int			idx, score;
    BUF			buf;

    msg_report("High Scores:\n");
    for (idx = 0; idx < theScoreFiles.entries(); idx++)
    {
	buf.sprintf("%s: %d\n", theScoreFiles(idx).buffer(), theScoreValues(idx));
	msg_report(buf);
    }

    if (!idx)
    {
	msg_report("No maps won yet.\n");
    }
}

void
hiscore_addscore(BUF mapname, int score)
{
    int		idx;

    for (idx = 0; idx < theScoreFiles.entries(); idx++)
    {
	if (!strcmp(mapname.buffer(), theScoreFiles(idx).buffer()))
	{
	    // Ignore worse scores.
	    if (theScoreValues(idx) > score)
		return;
	    
	    // Update in place
	    theScoreValues.set(idx, score);
	    hiscore_save();
	    return;
	}
    }

    // Append
    theScoreFiles.append(mapname);
    theScoreValues.append(score);
    hiscore_save();
}

int
hiscore_getscore(BUF mapname)
{
    int		idx;

    for (idx = 0; idx < theScoreFiles.entries(); idx++)
    {
	if (!strcmp(mapname.buffer(), theScoreFiles(idx).buffer()))
	{
	    return theScoreValues(idx);
	}
    }
    return -1;
}
