/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        item.h ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#ifndef __item__
#define __item__

#include "glbdef.h"

#include "pos.h"
#include "grammar.h"
#include "ptrlist.h"

#include <iostream>
using namespace std;

class MAP;

class ITEM
{
public:
		~ITEM();

    static ITEM *create(ITEM_NAMES item, int depth = 0);

    static ITEM *createRandom(int depth);
    static ITEM_NAMES itemFromHash(unsigned hash);

    ITEM	*copy() const;

    ITEM_NAMES	 getDefinition() const { return myDefinition; }
		
    VERB_PERSON	 getPerson() const;
    BUF		 getName(bool usearticle = false) const;

    // Returns 0 if no detailed description.  Returns a multi-line
    // buffer prefixed with +-
    const char  *getDetailedDescription() const;

    const ITEM_DEF	&defn() const { return defn(getDefinition()); }
    static const ITEM_DEF &defn(ITEM_NAMES item) { return glb_itemdefs[item]; }

    SIZE_NAMES		 getSize() const { return (SIZE_NAMES) defn().size; }

    void	 getLook(u8 &symbol, ATTR_NAMES &attr) const;

    int		 getX() const { return myX; }
    int		 getY() const { return myY; }
    POS		 getPos() const { return POS(myX, myY); }

    bool	 isFinalPos() const { return myFinalPos; }
    void	 setFinalPos(bool val) { myFinalPos = val; }

    bool	 canStackWith(const ITEM *stack) const;
    // Merges item's properties with our own.  item should likely
    // be deleted afterwards.
    void	 combineItem(const ITEM *item);
    int		 getStackCount() const { return myCount; }
    void	 decStackCount() { myCount--; }

    // Separates off the given number of items into a new item
    // stack.  Error if either stack is then size 0.
    ITEM	*splitStack(int newstack);

    // -1 for things without a count down.
    int		 getTimer() const { return myTimer; }

    // Returns true if should self destruct.
    bool	 runHeartbeat();

    DPDF	 getMeleeDPDF() const;

    void	 getWeaponStats(int &power, int &accuracy, int &consistency) const;

    DPDF	 getRangeDPDF() const;
    int		 getRangeRange() const;
    int		 getRangeMana() const;
    int		 getRangeArea() const;

    void	 getRangeStats(int &range, int &power, int &consistency,
				int &area, int &mana) const;

    void	 move(MAP *map, int x, int y);

    // Used by map & inventory...
    ITEM	*getNext() const { return myNext; }
    void	 setNext(ITEM *next) { myNext = next; }

    void	 save(ostream &os) const;
    static ITEM	*load(istream &is);

protected:
		 ITEM();

    ITEM_NAMES	 myDefinition;

    int		 myX, myY;
    int		 myCount;
    int		 myTimer;
    bool	 myFinalPos;

    ITEM	*myNext;
};

typedef PTRLIST<ITEM *> ITEMLIST;

#endif

