/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        main.cpp ( Fatherhood Library, C++ )
 *
 * COMMENTS:
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>

#include "gfxengine.h"
#include "glbdef.h"
#include "map.h"
#include "mob.h"
#include "panel.h"
#include "msg.h"
#include "text.h"
#include "speed.h"
#include "item.h"
#include "score.h"
#include "hiscore.h"
#include "dpdf.h"
#include "avatar.h"
#include "dircontrol.h"

#include <fstream>
using namespace std;

#define MAP_DISPW 50
#define MAP_DISPH 28

// Might as well make these global and save some pain.
PANEL	glb_output(79 - MAP_DISPW - 2, 28-9, true);
PANEL	glb_status(79 - MAP_DISPW, 8);
PANEL	glb_popup(40, 15);
bool	glbPopupActive = false;

void fullRedraw(MAP *map=0);

void
showPopupPanel()
{
    msg_registerpanel(&glb_popup);
    gfx_clearbox(19, 7, 42, 17, ' ', ATTR_POPUPBORDER);
    glbPopupActive = true;

    // Erase teh contents of the popup, it is confusing to see your
    // last choice there.
    glb_popup.clear();

    fullRedraw();
}

void
hidePopupPanel()
{
    msg_registerpanel(&glb_output);
    glbPopupActive = false;
    gfx_clearbox(0, 0, 80, 30, ' ', ATTR_BORDER);
    // Want the map blank by default.
    // This way if intro text is long people don't see a big blue text.
    gfx_clearbox(1, 1, MAP_DISPW, MAP_DISPH, ' ', ATTR_NORMAL);
    fullRedraw();
}

void
displayMapName(MAP *map)
{
    // Clear out old name.
    gfx_clearbox(0, 0, 80, 1, ' ', ATTR_BORDER);

    // Write in new name centered.
    BUF		name;

    name = map->getMapName();
    if (!name.isstring())
	name.reference("Random Map");

    int		x;
    x = 1 + MAP_DISPW/2;
    x -= name.strlen()/2;
    if (x < 0)
	x = 0;
    gfx_printstring(x, 0, name.buffer(), ATTR_BORDER);
}

void
save_stuff(MAP *mainmap)
{
    // Did I mention my hatred of streams?
#ifdef WIN32
    ofstream	os("fatherhood.sav", ios::out | ios::binary);
#else
    ofstream	os("fatherhood.sav");
#endif

    score_save(os);
    avatar_save(os);

    MAP		*map;
    u8		 c;

    for (map = mainmap; map; map = map->peekDownMap())
    {
	c = 1;
	os.write((const char *) &c, 1);
	map->save(os);
    }
    c = 0;
    os.write((const char *) &c, 1);
}

MAP *
load_stuff()
{
    // Destroy any existing maps...
    MOB::setAvatar(0);

    // Open file for reading.
#ifdef WIN32
    ifstream	is("fatherhood.sav", ios::in | ios::binary);
#else
    ifstream	is("fatherhood.sav");
#endif

    if (!is)
	return 0;

    u8			c;
    
    score_load(is);
    avatar_load(is);

    // Load the maps...
    MAP			*map, *lastmap, *mainmap;

    lastmap = 0;
    while (1)
    {
	is.read((char *) &c, 1);
	if (!c)
	    break;

	map = new MAP;
	map->load(is);

	if (lastmap)
	{
	    lastmap->setDownMap(map);
	    map->setUpMap(lastmap);
	}
	else
	    mainmap = map;

	lastmap = map;
    }

    // Find the avatar...
    mainmap->locateAvatar(MOB_AVATAR);

    msg_report(text_lookup("welcomeback", "You"));

    msg_newturn();
    
    return mainmap;
}

bool
getYes(bool preferyes)
{
    fullRedraw();

    if (preferyes)
    {
	switch (gfx_getKey())
	{
	    case 'n':
	    case 'N':
		return false;
		break;
	}
	return true;
    }
    else
    {
	switch (gfx_getKey())
	{
	    case 'y':
	    case 'Y':
		return true;
		break;
	}
    }
    return false;
}


void
writemobstatus(MOB_NAMES name, MAP *map, MOB *mob, int y)
{
    BUF		 buf;
    ITEM	*item;
    if (!mob)
    {
	if (name == MOB_AVATAR)
	    buf.reference("You are dead.");
	else
	{
	    if (map && map->everHad(name))
		buf.sprintf("%s is dead.", glb_mobdefs[name].name);
	}
	    
	glb_status.fillRect(0, y, 35, 1, ATTR_NORMAL);

	glb_status.appendText(buf);
	glb_status.newLine();
    }
    else
    {
	BUF	itemname;
	
	if (!mob->isAvatar())
	{
	    buf.sprintf("%s, %s", glb_mobdefs[name].name, glb_activitydefs[mob->getActivity()].name);
	    glb_status.fillRect(0, y, 35, 1, (ATTR_NAMES)glb_activitydefs[mob->getActivity()].attr);
	}
	else
	    buf.sprintf("%s", glb_mobdefs[name].name);
	glb_status.appendText(buf);
	glb_status.newLine();
	
	item = mob->getInventory();
	itemname.reference("nothing");
	if (item)
	    itemname = item->getName(true);

	buf.sprintf("  Carrying %s", itemname.buffer());
	glb_status.appendText(buf);
    }
}

void
writestatusline(MAP *mainmap)
{
    MOB		*avatar, *child = 0;
    int		 y = 0;

    avatar = MOB::getAvatar();

    glb_status.clear();

    writemobstatus(MOB_AVATAR, mainmap, avatar, y);
    y += 2;
    glb_status.newLine();

    child = mainmap->findMob(MOB_PRIMUS);
    writemobstatus(MOB_PRIMUS, mainmap, child, y);
    y += 2;
    glb_status.newLine();

    child = mainmap->findMob(MOB_DEUCE);
    writemobstatus(MOB_DEUCE, mainmap, child, y);
    y += 2;
    glb_status.newLine();

    child = mainmap->findMob(MOB_TREO);
    writemobstatus(MOB_TREO, mainmap, child, y);
    y += 2;
}

// Contrary to its name, it redraws everything EXCEPT the map!
void
fullRedraw(MAP *mainmap)
{
    if (!mainmap)
    {
	if (MOB::getAvatar())
	{
	    mainmap = MOB::getAvatar()->getMap();
	}
	
    }

    if (glbPopupActive)
	glb_popup.redraw();
    else
    {
	if (mainmap)
	    writestatusline(mainmap);
	glb_output.redraw();
	glb_status.redraw();
    }

    gfx_update();
}

// Rebuilds the avatar's FOV and its map.
// Does not do gfx update!
void
redrawAvatarMap()
{
    MOB		*avatar;

    avatar = MOB::getAvatar();
    if (!avatar)
	return;
    avatar->getMap()->buildFOV(avatar->getX(), avatar->getY());
    avatar->getMap()->redraw(1, 1, MAP_DISPW, MAP_DISPH, avatar->getX(), avatar->getY());
}
    

bool
getdirection(int key, int &dx, int &dy, bool allowfast = false)
{
    dx = dy = 0;
    bool isfast = false;

    switch (key)
    {
	case 'Y':
	    isfast = true;
	case 'y':
	case '7':
	    dx = -1;
	    dy = -1;
	    break;
	case 'U':
	    isfast = true;
	case 'u':
	case '9':
	    dx = 1;
	    dy = -1;
	    break;
	case 'H':
	    isfast = true;
	case '4':
	case 'h':
	case GFX_KEYLEFT:
	    dx = -1;
	    dy = 0;
	    break;
	case 'J':
	    isfast = true;
	case '2':
	case 'j':
	case GFX_KEYDOWN:
	    dx = 0;
	    dy = 1;
	    break;
	case 'K':
	    isfast = true;
	case '8':
	case 'k':
	case GFX_KEYUP:
	    dx = 0;
	    dy = -1;
	    break;
	case 'L':
	    isfast = true;
	case '6':
	case 'l':
	case GFX_KEYRIGHT:
	    dx = 1;
	    dy = 0;
	    break;
	case 'B':
	    isfast = true;
	case 'b':
	case '1':
	    dx = -1;
	    dy = 1;
	    break;
	case 'N':
	    isfast = true;
	case 'n':
	case '3':
	    dx = 1;
	    dy = 1;
	    break;
	case '5':
	case '.':
	    dx = 0;
	    dy = 0;
	    break;
	    
	default:
	    // No valid direction!
	    return false;
    }

    if (isfast && !allowfast)
	return false;
    if (isfast)
    {
	dx *= 5;
	dy *= 5;
    }

    return true;
}

bool
selectsquare(int &x, int &y, PANEL &output, PANEL &status)
{
    bool		selected;
    bool		isblind;
    int			key, dx, dy;
    MOB			*avatar;

    avatar = MOB::getAvatar();
    if (!avatar)
	return false;

    isblind = avatar->isBlind();
    
    selected = false;
    // Examine...
    while (1)
    {
	// Highlight current position
	avatar->getMap()->redraw(1, 1, MAP_DISPW, MAP_DISPH, x, y);

	u8		sym;
	ATTR_NAMES  attr;

	gfx_getchar(MAP_DISPW/2+1, MAP_DISPH/2+1, sym, attr);
	attr = ATTR_HILITE;
	gfx_printchar(MAP_DISPW/2+1, 15, sym, attr);

	fullRedraw();
	
	// Get a direction & apply it.
	key = gfx_getKey();
	
	// Exit if bad...
	if (!getdirection(key, dx, dy, true))
	{
	    if (key == ' ' || key == '\n' || key == '\r')
	    {
		selected = true;
		break;
	    }
	    else if (key == '>')
	    {
		int		tx, ty;
		if (avatar->getMap()->findTile(TILE_DOWNSTAIRS, tx, ty))
		{
		    dx = tx - x;
		    dy = ty - y;
		}
	    }
	    else if (key == '<')
	    {
		int		tx, ty;
		if (avatar->getMap()->findTile(TILE_UPSTAIRS, tx, ty))
		{
		    dx = tx - x;
		    dy = ty - y;
		}
	    }
	    else if (key == '/')
	    {
		int		tx, ty;
		
		// Search for map symbol...
		msg_report("Jump to what symbol?  ");
		fullRedraw();
		key = gfx_getKey();
		tx = x; 
		ty = y;
		if (avatar->getMap()->findNextSymbol(key, tx, ty))
		{
		    msg_report("Done.  ");
		    dx = tx - x;
		    dy = ty - y;
		}
		else
		{
		    msg_report("Not found.  ");
		}
	    }
	    else
	    {
		break;
	    }
	}

	x += dx;
	y += dy;

	if (!avatar->getMap()->isValidCoord(x, y))
	{
	    if (x < 0)
		x = 0;
	    if (x >= avatar->getMap()->getWidth())
		x = avatar->getMap()->getWidth()-1;
	    if (y < 0)
		y = 0;
	    if (y >= avatar->getMap()->getHeight())
		y = avatar->getMap()->getHeight()-1;
	}

	avatar->getMap()->describeSquare(x, y, isblind);
	
	msg_newturn();
    }

    return selected;
}

MOB *
createAvatar(MAP *mainmap)
{
    // Assume our map has created ourselves...
    mainmap->locateAvatar(MOB_AVATAR);

    if (!MOB::getAvatar())
    {
	int		x, y;
	// Oops. build him
	MOB *avatar;
	avatar = MOB::create(MOB_AVATAR);
	MOB::setAvatar(avatar);
	mainmap->findMoveable(x, y, true);
	avatar->move(x, y);
	mainmap->addMob(avatar);
    }

    return MOB::getAvatar();
}

MAP *
buildRandomWorld(MAP *mainmap)
{
    MAPSIZE_NAMES	size;
    BUF			buf;
    char		input[100];
    int			nriver, nfire;

    msg_report("What size?\n");
    FOREACH_MAPSIZE(size)
    {
	buf.sprintf("%d) %s\n", size, glb_mapsizedefs[size].name);
	msg_report(buf);
    }

    msg_getString("> ", input, 30);
    size = (MAPSIZE_NAMES) atoi(input);
    if (size < 0) size = (MAPSIZE_NAMES) (0);
    if (size >= NUM_MAPSIZES) size = (MAPSIZE_NAMES) (NUM_MAPSIZES-1);

    buf.sprintf("Building a %s map.\n", glb_mapsizedefs[size].name);
    msg_report(buf);

    msg_report("How many rivers?\n");
    msg_getString("> ", input, 30);
    nriver = atoi(input);

    msg_report("How many fires?\n");
    msg_getString("> ", input, 30);
    nfire = atoi(input);

    buf.sprintf("%d rivers and %d fires selected.\n", nriver, nfire);
    msg_report(buf);
    msg_report("Building: ");
    fullRedraw();

    do
    {
	// Delete the old map.
	delete mainmap;

	mainmap = new MAP;
	// build our world
	mainmap->build(size, nriver, nfire);
    } while (mainmap->getRooms() < 20);

    return mainmap;
}

MAP *
loadExistingWorld(MAP *mainmap)
{
    BUF			buf;
    char		input[100];
    const char		*f;
    int			i;
    PTRLIST<BUF>	fnames;
    DIRECTORY		dir;

    dir.opendir("../rooms");
    
    while (f = dir.readdir())
    {
	buf.strcpy(f);
	if (buf.extensionMatch("map"))
	    fnames.append(buf);
    }

    if (!fnames.entries())
    {
	msg_report("No valid .map files found!\n");
	return buildRandomWorld(mainmap);
    }

    msg_report("Which file?\n");
    for (i = 0; i < fnames.entries(); i++)
    {
	buf.sprintf("%d) %s\n", i, fnames(i).buffer());
	msg_report(buf);
    }

    msg_getString("> ", input, 30);
    i = atoi(input);

    if (i < 0) i = 0;
    if (i >= fnames.entries()) i = fnames.entries()-1;

    buf.sprintf("Loading %s.\n", fnames(i).buffer());
    msg_report(buf);

    // Delete the old map.
    delete mainmap;

    mainmap = new MAP;
    // Load our starting world
    buf.sprintf("../rooms/%s", fnames(i).buffer());
    mainmap->loadMap(buf.buffer());
    mainmap->setMapName(fnames(i).buffer());

    return mainmap;
}

MAP *
newGame(MAP *mainmap)
{
    MOB		*avatar;

    fullRedraw(mainmap);

    showPopupPanel();

    score_init();

    msg_report("\nPrebuilt World? (Y/n)\n");
    if (getYes(true))
	mainmap = loadExistingWorld(mainmap);
    else
	mainmap = buildRandomWorld(mainmap);

    // Create a new avatar.
    avatar = createAvatar(mainmap);

    hidePopupPanel();

    displayMapName(mainmap);

    if (mainmap->getIntroText().isstring())
	msg_report(mainmap->getIntroText());
    else
	msg_report(text_lookup("welcome", avatar->getRawName()));

    mainmap->mapConnected(avatar->getPos());

    redrawAvatarMap();
    fullRedraw();

    msg_newturn();

    return mainmap;
}

int
main()
{
    gfx_init();
    text_init();
    spd_init();
    score_init();

    rand_setseed((long) time(0));

    MAP		*mainmap;
    int		key;
    MOB		*avatar;

    hiscore_load();

    gfx_clearbox(0, 0, 80, 30, ' ', ATTR_BORDER);
    // Want the map blank by default.
    gfx_clearbox(1, 1, MAP_DISPW, MAP_DISPH, ' ', ATTR_NORMAL);

    glb_output.move(MAP_DISPW+2, 10);
    glb_status.move(MAP_DISPW+2, 1);
    glb_popup.move(20, 8);

    // Scroll to bottom of output...
    {
	int		i;
	for (i = 0; i < 30; i++)
	    glb_output.newLine();
    }

    msg_registerpanel(&glb_output);

    int		dx, dy;
    bool	didmove;
    bool	forcemove;
    bool	hasquit = false, newturn = false;
    bool	autowalkon = false;

    if (mainmap = load_stuff())
    {
	// On successful load, delete the save game.
	unlink("fatherhood.sav");
    }
    else
    {
	// start the game!
	mainmap = newGame(0);
    }
    displayMapName(mainmap);

    // Create a new avatar.
    avatar = MOB::getAvatar();

    while (!hasquit)
    {
	avatar = MOB::getAvatar();

	if (!avatar || avatar_victory())
	{
	    // Quit the game!
	    msg_report("\n");

	    if (avatar_victory())
		msg_report(text_lookup("game", "victory"));
	    else
		msg_report(text_lookup("game", "lose"));

	    // Give the chance to go back...
	    msg_report("Hit Q to Start Again.");

	    while (1)
	    {
		fullRedraw(mainmap);
		mainmap->re_redraw();

		key = gfx_getKey();

		if (key == 'Q')
		    break;

		spd_inctime();
		mainmap->doMoveNPC();

		if (mainmap->flowLiquids())
		{
		    BUF		buf;
		    buf.sprintf("The waters are dammed!  Your score: %d.\n", mainmap->calcScore());
		    msg_report(buf);
		}
		mainmap->drownMobs();
	    }
	    
	    // start the game!
	    mainmap = newGame(mainmap);

	    avatar = MOB::getAvatar();

	    autowalkon = false;
	}

	// avatar now guaranteed valid.

	didmove = false;
	forcemove = false;

	redrawAvatarMap();

	fullRedraw();

	// Check for force move.
	didmove = false;
	if (newturn)
	    forcemove = avatar->aiForcedAction();
	if (forcemove)
	    didmove = true;

	// Status line may trigger an update, due to losing shields.
	fullRedraw();

	// Avatar could die in forced action
	avatar = MOB::getAvatar();
	if (!avatar)
	    continue;

	// Cancel autowalk if key pressed.
	if (autowalkon && gfx_isKeyWaiting())
	{
	    // Eat the key
	    gfx_getKey();
	    msg_report("Interrupted.  ");
	    autowalkon = false;
	}

	if (!didmove && autowalkon)
	{
	    if (avatar->isSwallowed())
	    {
		msg_report("There is no where to run in here.  ");
	    }
	    else
	    {
		didmove = avatar->aiPathFindTo(avatar->getHX(), avatar->getHY());
		fullRedraw();
		gfx_update();
	    }
	    // Failed to path find, abort the walk.
	    if (!didmove)
		autowalkon = false;
	}
	
	if (!didmove)
	    key = gfx_getKey();
	else
	    key = 0;

	// Pre-empt usual keypresses with explore mode presses.
	if (getdirection(key, dx, dy))
	{
	    // Move in the requested direction, reset the key
	    key = 0;
	    didmove = avatar->actionBump(dx, dy);
	}

	if (key == '<' || key == '>')
	{
	    didmove = avatar->actionClimb();
	    key = 0;
	}

	switch (key)
	{
	    case ' ':
		didmove = avatar->actionBump(0, 0);
		msg_newturn();
		break;

	    case 'w':
	    {
		// Setup the auto walk.
		int		tx = avatar->getX(), ty = avatar->getY();

		msg_report("Autowalk to where?  ");
		fullRedraw();
		gfx_update();
		if (selectsquare(tx, ty, glb_output, glb_status))
		{
		    if (avatar->isSwallowed())
			msg_report("There is no where to run in here.  ");
		    else
		    {
			msg_report("Engaged.  ");
			avatar->setHome(tx, ty);
			autowalkon = true;
		    }
		}
		else
		{
		    msg_report("Cancelled.  ");
		}
		break;
	    }

	    case 'S':
		hiscore_display();
		msg_newturn();
		break;

	    case 'o':
	    {
		char		input[100];

		msg_report("Save Map File of?\n");
		msg_report("0) Current Map\n");
		msg_report("1) Initial Map\n");

		msg_getString("> ", input, 30);
		switch (*input)
		{
		    case '0':
		    case '1':
		    {
			BUF		buf, fname;

			if (*input == '0')
			    buf = mainmap->saveToText();
			else
			    buf = mainmap->getInitialText();

			msg_report("Save as?\n");
			msg_getString("> ", input, 30);

			if (!*input)
			{
			    msg_report("You decide not to save.\n");
			}
			else
			{
			    BUF		inputname;

			    inputname.reference(input);
			    inputname.makeFileSafe();
			    if (!inputname.extensionMatch("map"))
				inputname.strcat(".map");

			    mainmap->setMapName(inputname.buffer());
			    displayMapName(mainmap);
			    
			    fname.sprintf("../rooms/%s", inputname.buffer());

			    ofstream	os(fname.buffer());

			    os.write(buf.buffer(), buf.strlen());
			}

			break;
		    }
		    default:
			msg_report("You decide not to save.\n");
			break;
		}
		msg_newturn();

		break;
	    }

	    case 'v':
		// Verify that they have won.
		if (mainmap->flowLiquids(true))
		{
		    BUF		buf;
		    int		score = mainmap->calcScore();
		    buf.sprintf("The waters are dammed!  Your score: %d.\n", score);
		    msg_report(buf);

		    msg_report("Declare victory? (Y/n)\n");
		    if (getYes(true))
		    {
			msg_report(text_lookup("game", "victory"));

			BUF		mapname;

			// Check for high score.
			mapname = mainmap->getMapName();
			if (!mapname.isstring())
			{
			    msg_report("Random maps are scored only if they are 'o'utput.\n");
			}
			else
			{
			    int		hiscore = hiscore_getscore(mapname);

			    if (hiscore > score)
			    {
				buf.sprintf("You did not beat the high score of %d.\n", hiscore);
				msg_report(buf);
			    }
			    else if (hiscore < 0)
			    {
				msg_report("Yours is the first win for this map!\n");
				hiscore_addscore(mapname, score);
			    }
			    else
			    {
				buf.sprintf("You beat the high score of %d!\n", hiscore);
				msg_report(buf);
				hiscore_addscore(mapname, score);
			    }
			}

			// Give chance to admire handywork
			msg_report("Hit Q to Start Again.");

			while (1)
			{
			    fullRedraw(mainmap);
			    mainmap->re_redraw();

			    key = gfx_getKey();

			    if (key == 'Q')
				break;

			    spd_inctime();
			    mainmap->doMoveNPC();
			}
			
			mainmap = newGame(mainmap);

			avatar = MOB::getAvatar();
		    }
		}
		else
		{
		    BUF		buf;
		    buf.sprintf("The waters still flows!  Current score: %d.\n", avatar->getMap()->calcScore());
		    msg_report(buf);
		    
		    msg_report("Wait for water? (y/N)\n");
		    if (getYes(false))
		    {
			bool		done = false;
			while (!done)
			{
			    done = mainmap->flowLiquids(true);
			    mainmap->drownMobs(true);
			    fullRedraw(mainmap);
			    mainmap->re_redraw();
			    if (gfx_isKeyWaiting())
			    {
				// Eat the key
				gfx_getKey();
				msg_report("Interrupted.  ");
				break;
			    }
			}
			if (done)
			    msg_report("The waters are dammed.");
		    }
		    else
		    {
			msg_report("Give up? (y/N)\n");

			if (getYes(false))
			{
			    msg_report(text_lookup("game", "lose"));
			    mainmap = newGame(mainmap);

			    avatar = MOB::getAvatar();
			}
		    }
		}
		msg_newturn();
		break;

	    case '+':
	    {
		// Fast forward time..
		int		tick;
		MAP		*map = avatar->getMap();

		msg_report("You stop and watch the water...");

		for (tick = 0; tick < 50; tick++)
		{
		    spd_inctime();
		    map->flowLiquids();
		    map->drownMobs();
		}
		msg_newturn();
		break;
	    }
		    
	    case 'z':
	    {
		msg_report(text_lookup("welcome", avatar->getRawName()));
		msg_newturn();
		break;
	    }

	    case ',':
	    case 'g':
		didmove = avatar->actionPickup();
		msg_newturn();
		break;
	    
	    case 'd':
		// Set the final pos flag.
		if (avatar->getInventory())
		    avatar->getInventory()->setFinalPos(true);
		didmove = avatar->actionDrop();
		msg_newturn();
		break;

	    case 'm':
		// Mark all rocks here as playable.
		didmove = avatar->actionMarkFree();
		msg_newturn();
		break;

	    case 't':
	    case 's':
	    case 'p':
	    case 'r':
	    {
		int		k;
		MOB		*child;
		bool		validchoice = true;
		MOB_NAMES	cname;
		BUF		buf;

		switch (key)
		{
		    case 't':
			msg_report("Talk to who? (1-3)\n");
			break;
		    case 's':
			msg_report("Scold who? (1-3)\n");
			break;
		    case 'p':
			msg_report("Praise who? (1-3)\n");
			break;
		    case 'r':
			msg_report("Recall who? (1-3)\n");
			break;
		}

		fullRedraw();
		k = gfx_getKey();
		switch (k)
		{
		    case '1':
			cname = MOB_PRIMUS;
			break;
		    case '2':
			cname = MOB_DEUCE;
			break;
		    case '3':
			cname = MOB_TREO;
			break;
		    default:
			validchoice = false;
			break;
		}

		// If invalid choice made, abandon.
		if (!validchoice)
		{
		    msg_report("You decide to keep quiet.\n");
		    msg_newturn();
		    break;
		}
		child = avatar->getMap()->findMob(cname);

		// If the child is dead...
		if (!child)
		{
		    buf.sprintf("You start to speak but then you freeze.  %s is dead!  ", glb_mobdefs[cname].name);
		    msg_report(buf);
		    msg_newturn();
		    break;
		}

		switch (key)
		{
		    case 'p':
			didmove = child->actionPraised();
			break;
		    case 's':
			didmove = child->actionScolded();
			break;
		    case 'r':
			didmove = child->actionRecalled();
			break;
		    case 't':
		    {
			char	speechify[100];

			msg_getString("> ", speechify, 30);
			msg_format("%S <say>: ", avatar);
			buf.sprintf("\"%s\"\n", speechify);
			msg_report(buf);
			didmove = true;
			break;
		    }
		    default:
			msg_report("You keep quiet.");
			break;
		}
		msg_newturn();

		break;
	    }
		
		
	    case 'i':
	    {
		BUF		buf;

		msg_report("You are carrying:\n");

		ITEM		*inv;

		inv = avatar->getInventory();
		if (inv)
		{
		    while (inv)
		    {
			msg_format("  %S\n", inv);
			// Write detailed information about weapons
			// and spellbooks.
			{
			    const char *buf;

			    buf = inv->getDetailedDescription();

			    if (buf)
			    {
				msg_report(buf);
			    }
			}
			inv = inv->getNext();
		    }
		}
		else
		{
		    msg_report("  Nothing.\n");
		}
		msg_newturn();

		break;
	    }

	    case 'Q':
	    {
		msg_report("Saving...  ");

		fullRedraw();

		save_stuff(mainmap);

		msg_report("Done.  Hit a key to quit.  ");
		
		fullRedraw();

		gfx_breakKeyRepeat();
		gfx_clearKeyBuffer();

		key = gfx_getKey();
		key = 0;
		hasquit = true;
		break;
	    }

	    case 'x':
	    {
		int		x, y;
		bool		selected;
		
		x = avatar->getX();
		y = avatar->getY();

		selected = selectsquare(x, y, glb_output, glb_status);
		
		if (selected)
		{
		    // If there is a mob there print its info.
		    MOB		*m;

		    m = avatar->getMap()->getMob(x, y);

		    if (m && !avatar->isBlind())
		    {
			m->viewDescription();
		    }
		}
		msg_newturn();
		
		// Screen restored for us...
		break;
	    }

	    case '?':
		msg_report("Help:\n");
		msg_report(text_lookup("game", "help"));
		msg_newturn();
		break;

	    case 'a':
		msg_report("About:\n");
		msg_report(text_lookup("game", "about"));
		msg_newturn();
		break;
	}

	// IFF didmove is set do we update all the NPCs.
	if (didmove)
	{
	    mainmap->doMoveNPC();

	    mainmap->flowLiquids();
	    mainmap->drownMobs();

	    spd_inctime();
	    newturn = true;

	    if (!forcemove)
	    {
		msg_newturn();
	    }
	}
	else
	    newturn = false;
    }

    text_shutdown();
    gfx_shutdown();

    return 0;
}
