/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        mapdata.cpp ( Save Sucmmer Library, C++ )
 *
 * COMMENTS:
 */

#include "mapdata.h"

#include <memory.h>
#include <assert.h>

class MAPDATA_REF
{
public:
    void	 incRef();
    void	 decRef();
    
    // Ensure I am the only pointer to the data.
    MAPDATA_REF	*uniquify();

    int		*data() { return myData; }

    // Returns a ref that already has been inced!
    static MAPDATA_REF *create(int nbytes);

private:
		 MAPDATA_REF();
    virtual	~MAPDATA_REF();

    int		 myRef;
    int		 mySize;
    int		*myData;
};

MAPDATA_REF::MAPDATA_REF()
{
    myData = 0;
    mySize = 0;
    myRef = 0;
}

MAPDATA_REF::~MAPDATA_REF()
{
    delete [] myData;
}

void
MAPDATA_REF::incRef()
{
    myRef++;
}

void
MAPDATA_REF::decRef()
{
    myRef--;
    if (myRef <= 0)
	delete this;
}

MAPDATA_REF *
MAPDATA_REF::uniquify()
{
    // Empty refs are easy to unique :>
    if (!myData)
	return this;

    // if only one ref, it is trivial
    if (myRef <= 1)
	return this;

    // Need to copy for upcoming write!
    MAPDATA_REF	*result;

    result = create(mySize);
    memcpy(result->data(), data(), mySize * sizeof(int));

    // Remove a ref from ourself.
    decRef();

    return result;
}

MAPDATA_REF *
MAPDATA_REF::create(int nbytes)
{
    MAPDATA_REF		*result;

    result = new MAPDATA_REF;
    result->incRef();
    result->myData = new int[nbytes];
    result->mySize = nbytes;

    return result;
}

//
// Mapdata functions
//

MAPDATA::MAPDATA()
{
    myRef = 0;
    myWidth = myHeight = 0;
}

MAPDATA::MAPDATA(int width, int height)
{
    myRef = MAPDATA_REF::create(width * height);
    myWidth = width;
    myHeight = height;

    // What everyone expects anyways.
    constant(0);
}

MAPDATA::MAPDATA(const MAPDATA &data)
{
    myRef = 0;

    *this = data;
}

MAPDATA::~MAPDATA()
{
    if (myRef)
	myRef->decRef();
}


MAPDATA &
MAPDATA::operator=(const MAPDATA &data)
{
    // Trivial assignment.
    if (this == &data)
	return *this;

    if (data.myRef)
	data.myRef->incRef();
    
    if (myRef)
	myRef->decRef();

    myRef = data.myRef;
    myWidth = data.myWidth;
    myHeight = data.myHeight;

    return *this;
}

int
MAPDATA::operator()(int x, int y) const
{
    assert(x >= 0 && x < myWidth);
    assert(y >= 0 && y < myHeight);
    return readData()[y * myWidth + x];
}

int &
MAPDATA::data(int x, int y)
{
    assert(x >= 0 && x < myWidth);
    assert(y >= 0 && y < myHeight);
    return writeData()[y * myWidth + x];
}

void
MAPDATA::constant(int value)
{
    int		i, n;
    if (!value)
	memset(writeData(), 0, myWidth * myHeight * sizeof(int));
    else
    {
	n = myWidth * myHeight;
	for (i = 0; i < n; i++)
	{
	    writeData()[i] = value;
	}
    }
}

void
MAPDATA::load(istream &is)
{
    is.read((char *) writeData(), myWidth * myHeight * sizeof(int));
}

void
MAPDATA::save(ostream &os) const
{
    os.write((const char *) readData(), myWidth * myHeight * sizeof(int));
}

const int *
MAPDATA::readData() const
{
    return myRef->data();
}

int *
MAPDATA::writeData()
{
    myRef = myRef->uniquify();
    return myRef->data();
}
